from yadm import fields
from yadm.documents import Document


def test_money_str():
    assert str(fields.Money('3.14')) == '3.14'
    assert str(fields.Money('0.14')) == '0.14'
    assert str(fields.Money('30.14')) == '30.14'
    assert str(fields.Money('3.1')) == '3.10'
    assert str(fields.Money('0.1')) == '0.10'


def test_money_to_mongo():
    assert fields.Money('3.14').to_mongo() == 314
    assert fields.Money('0.14').to_mongo() == 14
    assert fields.Money('30.14').to_mongo() == 3014
    assert fields.Money('3.1').to_mongo() == 310
    assert fields.Money('0.1').to_mongo() == 10


class TestDoc(Document):
    __collection__ = 'testdocs'
    money = fields.MoneyField()


def test_save(db):  # noqa
    doc = TestDoc()
    doc.money = fields.Money('3.14')
    db.save(doc)

    data = db.db.testdocs.find_one()

    assert 'money' in data
    assert isinstance(data['money'], int)
    assert data['money'] == 314


def test_load(db):
    db.db.testdocs.insert({'money': 314})

    doc = db.get_queryset(TestDoc).find_one()

    assert hasattr(doc, 'money')
    assert isinstance(doc.money, fields.Money)
    assert doc.money == fields.Money('3.14')


def test_load_and_save(db):
    db.db.testdocs.insert({'money': 314})

    doc = db.get_queryset(TestDoc).find_one()
    doc.money = fields.Money('42.00')
    db.save(doc)

    assert hasattr(doc, 'money')
    assert isinstance(doc.money, fields.Money)
    assert doc.money == fields.Money('42.00')

    data = db.db.testdocs.find_one()

    assert 'money' in data
    assert isinstance(data['money'], int)
    assert data['money'] == 4200



import pytest

import pymongo

from yadm.database import Database
from yadm.documents import Document
from yadm.queryset import QuerySet
from yadm.serialize import from_mongo
from yadm import fields


class TestDoc(Document):
    __collection__ = 'testdocs'
    b = fields.BooleanField()
    i = fields.IntegerField()
    l = fields.ListField(fields.IntegerField())


def test_init(db):
    assert isinstance(db.client, pymongo.mongo_client.MongoClient)
    assert isinstance(db.db, pymongo.database.Database)
    assert db.db.name == 'test'
    assert db.name == 'test'


def test_get_collection(db):
    collection = db._get_collection(TestDoc)

    assert isinstance(collection, pymongo.collection.Collection)
    assert collection.name == 'testdocs'


def test_get_queryset(db):
    queryset = db.get_queryset(TestDoc)

    assert isinstance(queryset, QuerySet)
    assert queryset._db is db
    assert queryset._document_class is TestDoc


def test_insert(db):
    doc = TestDoc()
    doc.i = 13

    db.insert(doc)

    assert db.db.testdocs.find().count() == 1
    assert db.db.testdocs.find()[0]['i'] == 13
    assert not doc.__changed__
    assert doc.__db__ is db


def test_save_new(db):
    doc = TestDoc()
    doc.i = 13

    db.save(doc)

    assert db.db.testdocs.find().count() == 1
    assert db.db.testdocs.find()[0]['i'] == 13
    assert not doc.__changed__
    assert doc.__db__ is db


def test_save(db):
    col = db.db.testdocs
    col.insert({'i': 13})

    doc = from_mongo(TestDoc, col.find_one({'i': 13}))
    doc.i = 26

    col.update({'_id': doc.id}, {'$set': {'b': True}})

    db.save(doc)

    assert doc.i == 26
    assert doc.b is True
    assert db.db.testdocs.find().count() == 1
    assert db.db.testdocs.find()[0]['i'] == 26
    assert db.db.testdocs.find()[0]['b'] is True
    assert not doc.__changed__
    assert doc.__db__ is db


def test_save_full(db):
    col = db.db.testdocs
    col.insert({'i': 13})

    doc = from_mongo(TestDoc, col.find_one({'i': 13}))
    doc.i = 26

    col.update({'_id': doc.id}, {'$set': {'b': True}})

    db.save(doc, full=True)

    assert doc.i == 26
    assert not hasattr(doc, 'b')
    assert db.db.testdocs.find().count() == 1
    assert db.db.testdocs.find()[0]['i'] == 26
    assert 'b' not in db.db.testdocs.find()[0]
    assert not doc.__changed__
    assert doc.__db__ is db


@pytest.mark.parametrize('unset', [['i'], {'i': True}])
def test_update_one(db, unset):
    col = db.db.testdocs
    _id = col.insert({'i': 13})

    doc = from_mongo(TestDoc, col.find_one(_id))
    db.update_one(doc, set={'b': True}, unset=unset)

    assert doc.b
    assert not hasattr(doc, 'i')


def test_update_one__inc(db):
    col = db.db.testdocs
    _id = col.insert({'i': 12})

    doc = from_mongo(TestDoc, col.find_one(_id))
    db.update_one(doc, inc={'i': 1})

    assert doc.i == 13
    assert not hasattr(doc, 'b')


@pytest.mark.parametrize('cmd, v, r', [
    ('push', 4, [1, 2, 3, 4]),
    ('pull', 2, [1, 3]),
])
def test_update_one__push_pull(db, cmd, v, r):
    col = db.db.testdocs
    _id = col.insert({'l': [1, 2, 3]})

    doc = from_mongo(TestDoc, col.find_one(_id))
    db.update_one(doc, **{cmd: {'l': v}})

    assert doc.l == r


def test_remove(db):
    col = db.db.testdocs
    col.insert({'i': 13})

    doc = from_mongo(TestDoc, col.find_one({'i': 13}))

    assert col.count() == 1
    db.remove(doc)
    assert col.count() == 0


def test_reload(db):
    doc = TestDoc()
    doc.i = 1
    db.insert(doc)

    db.db.testdocs.update({'_id': doc.id}, {'i': 2})

    assert doc.i == 1
    new = db.reload(doc)
    assert doc.i == 2
    assert doc is new


def test_reload_new_instance(db):
    doc = TestDoc()
    doc.i = 1
    db.insert(doc)

    db.db.testdocs.update({'_id': doc.id}, {'i': 2})

    assert doc.i == 1
    new = db.reload(doc, new_instance=True)
    assert doc.i == 1
    assert doc is not new
    assert new.i == 2


def test_read_preference__init(client, mongo_args):
    host, port, name = mongo_args
    client.drop_database(name)
    sp = pymongo.read_preferences.ReadPreference.SECONDARY_PREFERRED
    db = Database(client, name, read_preference=sp)
    assert db.db.read_preference == sp


def test_read_preference__get_collection(db):
    sp = pymongo.read_preferences.ReadPreference.SECONDARY_PREFERRED
    col = db._get_collection(TestDoc, read_preference=sp)
    assert col.read_preference == sp



"""
This module for provide work with MongoDB database.

.. code-block:: python

    import pymongo
    from yadm.database import Database

    from mydocs import Doc

    client = pymongo.MongoClient("localhost", 27017)
    db = Database(self.client, 'test')

    doc = Doc()
    db.insert(doc)

    doc.arg = 13
    db.save(doc)

    qs = db.get_queryset(Doc).find({'arg': {'$gt': 10}})
    for doc in qs:
        print(doc)
"""
import pymongo

from yadm.markers import AttributeNotSet
from yadm.aggregation import Aggregator
from yadm.queryset import QuerySet
from yadm.bulk import Bulk
from yadm.serialize import to_mongo


PYMONGO_VERSION = pymongo.version_tuple


class Database:
    """ Main object who provide work with database.

    :param pymongo.Client client: database connection
    :param str name: database name
    """
    def __init__(self, client, name, *, read_preference=None):
        self.client = client
        self.name = name

        if PYMONGO_VERSION < (3, 0):
            self.db = client[name]

            if read_preference is not None:
                self.db.read_preference = read_preference

        else:
            self.db = client.get_database(name, read_preference=read_preference)

    def __repr__(self):  # pragma: no cover
        return 'Database({!r})'.format(self.db)

    def __call__(self, document_class, *, cache=None):
        return self.get_queryset(document_class, cache=cache)

    def _get_collection(self, document_class, *, read_preference=None):
        """ Return pymongo collection for document class.
        """
        if PYMONGO_VERSION < (3, 0):
            collection = self.db[document_class.__collection__]

            if read_preference is not None:
                collection.read_preference = read_preference

            return collection

        else:
            return self.db.get_collection(document_class.__collection__,
                                          read_preference=read_preference)

    def get_queryset(self, document_class, *, cache=None):
        """ Return queryset for document class.

        :param document_class: :class:`yadm.documents.Document`
        :param cache: cache for share with other querysets

        This create instance of :class:`yadm.queryset.QuerySet`
        with presetted document's collection information.
        """
        return QuerySet(self, document_class, cache=cache)

    def aggregate(self, document_class, *, pipeline=None):
        """ Return aggregator for use aggregation framework.

        :param document_class: :class:`yadm.documents.Document`
        :param list pipeline: initial pipeline
        """
        return Aggregator(self, document_class, pipeline=None)

    def insert(self, document):
        """ Insert document to database.

        :param Document document: document instance for insert to database

        It's bind new document to database set
        :py:attr:`_id <yadm.documents.Document._id>`.
        """
        document.__db__ = self
        collection = self._get_collection(document.__class__)
        document._id = collection.insert(to_mongo(document))
        document.__changed_clear__()
        return document

    def save(self, document, full=False, upsert=False):
        """ Save document to database.

        :param Document document: document instance for save
        :param bool full: fully resave document
            (default: `False`)
        :param bool upsert: see documentation for MongoDB's `update`
            (default: `False`)

        If document has no `_id`
        :py:meth:`insert <Database.insert>` new document.
        """
        if hasattr(document, '_id'):
            document.__db__ = self

            if full:
                self._get_collection(document).update(
                    {'_id': document.id},
                    to_mongo(document),
                    upsert=upsert,
                    multi=False,
                )
                document.__changed_clear__()
            else:
                set_data = to_mongo(
                    document,
                    exclude=['_id'],
                    include=list(document.__changed__),
                )

                unset_data = [f for f, v in document.__changed__.items()
                              if v is AttributeNotSet]

                self.update_one(document, set=set_data, unset=unset_data)

            return document
        else:
            return self.insert(document)

    def update_one(self, document, reload=True, *,
                   set=None, unset=None, inc=None,
                   push=None, pull=None):  # TODO: extend
        """ Update one document.

        :param Document document: document instance for update
        :param bool reload: if True, reload document
        """
        update_data = {}

        if set:
            update_data['$set'] = set

        if unset:
            if isinstance(unset, dict):
                update_data['$unset'] = unset
            else:
                update_data['$unset'] = {f: True for f in unset}

        if inc:
            update_data['$inc'] = inc

        if push:
            update_data['$push'] = push

        if pull:
            update_data['$pull'] = pull

        if update_data:
            self._get_collection(document).update(
                {'_id': document.id},
                update_data,
                upsert=False,
                multi=False,
            )

        if reload:
            self.reload(document)

    def remove(self, document):
        """ Remove document from database.

        :param Document document: instance for remove from database
        """
        return self._get_collection(document.__class__).remove({'_id': document._id})

    def reload(self, document, new_instance=False):
        """ Reload document.

        :param Document document: instance for reload
        :param bool new_instance: if `True` return new instance of document,
            else change data in given document (default: `False`)
        """
        new = self.get_queryset(document.__class__).find_one(document.id)

        if new_instance:
            return new
        else:
            document.__raw__.clear()
            document.__raw__.update(new.__raw__)
            document.__cache__.clear()
            document.__changed__.clear()
            return document

    def bulk(self, document_class, ordered=False, raise_on_errors=True):
        """ Return Bulk.

        :param MetaDocument document_class: class of documents fo bulk
        :param bool ordered: create ordered bulk (default `False`)
        :param bool raise_on_errors: raise BulkWriteError exception
            if write errors (default `True`)

        Context manager:

            with db.bulk(Doc) as bulk:
                bulk.insert(doc_1)
                bulk.insert(doc_2)
        """
        return Bulk(self, document_class, ordered, raise_on_errors)



import pytest

from yadm.testing import create_fake
from yadm.documents import Document
from yadm import fields


class TestDocRef(Document):
    __collection__ = 'testdocs_ref'
    s = fields.StringField()


class TestDoc(Document):
    __collection__ = 'testdocs'
    ref = fields.ReferenceField(TestDocRef)
    i = fields.IntegerField()


def test_reference(db):
    doc = create_fake(TestDoc, db)
    assert isinstance(doc, TestDoc)
    assert isinstance(doc.ref, TestDocRef)
    assert doc.ref.id



import pytest

from pymongo.errors import BulkWriteError

from yadm import fields
from yadm.documents import Document
from yadm.bulk import Bulk


class TestDoc(Document):
    __collection__ = 'testdocs'

    i = fields.IntegerField()


@pytest.fixture
def index(db):
    return db.db.testdocs.ensure_index([('i', 1)], unique=True)


def test_create(db):
    bulk = db.bulk(TestDoc)
    assert isinstance(bulk, Bulk)


def test_insert_one(db):
    doc = TestDoc()
    doc.i = 1

    bulk = db.bulk(TestDoc)
    bulk.insert(doc)

    assert db.db.testdocs.count() == 0

    bulk.execute()

    assert bulk.result
    assert db.db.testdocs.count() == 1
    assert db.db.testdocs.find_one()['i'] == 1


def test_insert_many(db):
    bulk = db.bulk(TestDoc)

    for i in range(10):
        doc = TestDoc()
        doc.i = i
        bulk.insert(doc)

    assert db.db.testdocs.count() == 0
    bulk.execute()
    assert db.db.testdocs.count() == 10


def test_insert_type_error(db):
    class Doc(Document):
        pass

    bulk = db.bulk(TestDoc)

    with pytest.raises(TypeError):
        bulk.insert(Doc())


def test_context_manager(db):
    with db.bulk(TestDoc) as bulk:
        doc = TestDoc()
        doc.i = 1
        bulk.insert(doc)

    assert db.db.testdocs.count() == 1


def test_context_manager_error(db):
    with pytest.raises(RuntimeError):
        with db.bulk(TestDoc) as bulk:
            doc = TestDoc()
            doc.i = 1
            bulk.insert(doc)

            raise RuntimeError

    assert db.db.testdocs.count() == 0


def test_result(db):
    with db.bulk(TestDoc) as bulk:
        doc = TestDoc()
        doc.i = 1
        bulk.insert(doc)

    assert bulk.result.n_inserted == 1


def test_result_write_error(db, index):
    with db.bulk(TestDoc, raise_on_errors=False) as bulk:
        doc = TestDoc()
        doc.i = 1
        bulk.insert(doc)

        doc = TestDoc()
        doc.i = 2
        bulk.insert(doc)

        doc = TestDoc()
        doc.i = 1
        bulk.insert(doc)

    assert not bulk.result
    assert db.db.testdocs.count() == 2
    assert bulk.result.n_inserted == 2
    assert len(bulk.result.write_errors) == 1
    assert bulk.result.write_errors[0].document.i == 1


def test_result_write_error_raise(db, index):
    with pytest.raises(BulkWriteError):
        with db.bulk(TestDoc) as bulk:
            doc = TestDoc()
            doc.i = 1
            bulk.insert(doc)

            doc = TestDoc()
            doc.i = 2
            bulk.insert(doc)

            doc = TestDoc()
            doc.i = 1
            bulk.insert(doc)

    assert bulk.result.n_inserted == 2
    assert len(bulk.result.write_errors) == 1
    assert bulk.result.write_errors[0].document.i == 1


def test_result_write_error_ordered(db, index):
    with db.bulk(TestDoc, ordered=True, raise_on_errors=False) as bulk:
        for i in range(10):
            doc = TestDoc()
            doc.i = i
            bulk.insert(doc)

            doc = TestDoc()
            doc.i = 1
            bulk.insert(doc)

    assert db.db.testdocs.count() == 2
    assert bulk.result.n_inserted == 2
    assert len(bulk.result.write_errors) == 1
    assert bulk.result.write_errors[0].document.i == 1



import pytest
import pymongo
from bson import ObjectId

from yadm.documents import Document
from yadm import fields
from yadm.markers import NotLoaded


class Doc(Document):
    __collection__ = 'testdocs'
    i = fields.IntegerField
    s = fields.StringField


@pytest.fixture
def qs(db):
    for n in range(10):
        db.db.testdocs.insert({
            'i': n,
            's': 'str({})'.format(n),
        })

    return db.get_queryset(Doc)


def test_count(qs):
    qs = qs.find({'i': {'$gte': 6}})
    assert qs.count() == 4


def test_len(qs):
    qs = qs.find({'i': {'$gte': 6}})
    assert len(qs) == 4


def test_find_one(qs):
    doc = qs.find_one({'i': 7})
    assert isinstance(doc, Doc)
    assert hasattr(doc, 'i')
    assert doc.i == 7


def test_find_exc(qs):
    class SimpleException(Exception):
        pass

    with pytest.raises(SimpleException):
        qs.find_one({'i': 100500}, exc=SimpleException)


def test_find(qs):
    qs = qs.find({'i': {'$gte': 6}})
    assert len([d for d in qs]) == 4
    assert {d.i for d in qs} == {6, 7, 8, 9}


def test_find_with_collisium(qs):
    qs = qs.find({'i': {'$gt': 4}})
    qs = qs.find({'i': {'$lt': 6}})
    assert qs.count() == 1
    assert qs[0].i == 5


def test_update(db, qs):
    qs.find({'i': {'$gte': 6}}).update({'$set': {'s': 'test'}})

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 4

    for doc in db.db.testdocs.find({'i': {'$lt': 6}}):
        assert doc['s'] != 'test'
        assert doc['s'].startswith('str(')

    for doc in db.db.testdocs.find({'i': {'$gte': 6}}):
        assert doc['s'] == 'test'


def test_update_not_multi(db, qs):
    qs.find({'i': {'$gte': 6}}).update({'$set': {'s': 'test'}}, multi=False)

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1


def test_find_and_modify(db, qs):
    doc = qs({'i': 6}).find_and_modify({'$set': {'s': 'test'}})

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert isinstance(doc, Document)
    assert doc.i == 6
    assert doc.s == 'str(6)'


def test_find_and_modify_new(db, qs):
    doc = qs({'i': 6}).find_and_modify(
        {'$set': {'s': 'test'}}, new=True)

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert isinstance(doc, Document)
    assert doc.i == 6
    assert doc.s == 'test'


def test_find_and_modify_full_response(db, qs):
    result = qs({'i': 6}).find_and_modify(
        {'$set': {'s': 'test'}}, full_response=True)

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert 'value' in result
    assert isinstance(result['value'], Document)
    assert result['value'].i == 6
    assert result['value'].s == 'str(6)'


def test_find_and_modify_full_response_new(db, qs):
    result = qs({'i': 6}).find_and_modify(
        {'$set': {'s': 'test'}}, full_response=True, new=True)

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert 'value' in result
    assert isinstance(result['value'], Document)
    assert result['value'].i == 6
    assert result['value'].s == 'test'


def test_find_and_modify_sort(db, qs):
    qs = qs({'i': {'$lte': 6, '$gte': 4}}).sort(('s', -1))
    doc = qs.find_and_modify({'$set': {'s': 'test'}})

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert isinstance(doc, Document)
    assert doc.i == 6
    assert doc.s == 'str(6)'


def test_find_and_modify_not_found(qs):
    ret = qs({'i': 13}).find_and_modify({'$set': {'s': 'test'}})
    assert ret is None


def test_remove(db, qs):
    qs.find({'i': {'$gte': 6}}).remove()
    assert len([d for d in qs]) == 6
    assert {d.i for d in qs} == set(range(6))

    assert db.db.testdocs.count() == 6
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(6))


def test_remove__one(db, qs):
    qs.find({'i': {'$gte': 6}}).remove(multi=False)
    assert len([d for d in qs]) == 9
    assert db.db.testdocs.count() == 9


def test_sort(qs):
    qs = qs.find({'i': {'$gte': 6}}).sort(('i', -1))
    assert [d.i for d in qs] == [9, 8, 7, 6]

    qs = qs.find({'i': {'$gte': 6}}).sort(('i', 1))
    assert [d.i for d in qs] == [6, 7, 8, 9]


def test_fields(qs):
    doc = qs.fields('s').find_one({'i': 3})

    assert doc.s, 'str(3)'
    assert 'i' in doc.__raw__
    assert doc.__raw__['i'] is NotLoaded

    with pytest.raises(fields.NotLoadedError):
        doc.i


def test_fields_all(qs):
    doc = qs.fields('s').fields_all().find_one({'i': 3})
    assert 'i' in doc.__data__
    assert doc.__data__['i'] is 3
    assert doc.s == 'str(3)'
    assert doc.i == 3


def test_read_preference(db):
    sp = pymongo.read_preferences.ReadPreference.SECONDARY_PREFERRED
    qs = db(Doc).read_preference(sp)
    assert qs._collection.read_preference == sp


def test_with_id(db, qs):
    _id = db.db.testdocs.find_one({'i': 4}, {'_id': True})['_id']
    doc = qs.with_id(_id)
    assert doc.s == 'str(4)'
    assert doc.i == 4


def test_contains(qs):
    doc = qs.find_one({'i': 0})
    assert doc in qs
    assert doc not in qs.find({'i': {'$ne': 0}})


def test_distinct(db, qs):
    res = qs.distinct('i')
    assert isinstance(res, list)
    assert len(res) == len(set(res))

    db.insert(Doc(i=3))
    db.insert(Doc(i=4))
    db.insert(Doc(i=10))

    res = qs.distinct('i')
    assert len(res) == len(set(res))


def test_bulk(qs):
    bulk = qs.find({'i': {'$gte': 6}}).bulk()
    assert isinstance(bulk, dict)
    assert len(bulk), 4
    assert isinstance(list(bulk)[0], ObjectId)
    assert isinstance(list(bulk.values())[0], Doc)
    _id = list(bulk)[0]
    assert bulk[_id].id == _id
    assert {d.i for d in bulk.values()} == {6, 7, 8, 9}
